@@ -181,7 +181,7 @@
  * 7) Mark state 3 final because state 5 of source NFA is marked as final.
  *
  *
- * Portions Copyright (c) 1996-2016, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2018, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -450,7 +450,7 @@ struct TrgmPackedGraph
 	 * by color trigram number.
 	 */
 	int			colorTrigramsCount;
-	int		   *colorTrigramGroups;		/* array of size colorTrigramsCount */
+	int		   *colorTrigramGroups; /* array of size colorTrigramsCount */
 
 	/*
 	 * The states of the simplified NFA.  State number 0 is always initial
@@ -634,7 +634,7 @@ createTrgmNFAInternal(regex_t *regex, TrgmPackedGraph **graph,
  * Main entry point for evaluating a graph during index scanning.
  *
  * The check[] array is indexed by trigram number (in the array of simple
- * trigrams returned by createTrgmNFA), and holds TRUE for those trigrams
+ * trigrams returned by createTrgmNFA), and holds true for those trigrams
  * that are present in the index entry being checked.
  */
 bool
@@ -1451,7 +1451,7 @@ prefixContains(TrgmPrefix *prefix1, TrgmPrefix *prefix2)
  * Get vector of all color trigrams in graph and select which of them
  * to expand into simple trigrams.
  *
- * Returns TRUE if OK, FALSE if exhausted resource limits.
+ * Returns true if OK, false if exhausted resource limits.
  */
 static bool
 selectColorTrigrams(TrgmNFA *trgmNFA)
@@ -1935,9 +1935,7 @@ packGraph(TrgmNFA *trgmNFA, MemoryContext rcontext)
 				arcsCount;
 	HASH_SEQ_STATUS scan_status;
 	TrgmState  *state;
-	TrgmPackArcInfo *arcs,
-			   *p1,
-			   *p2;
+	TrgmPackArcInfo *arcs;
 	TrgmPackedArc *packedArcs;
 	TrgmPackedGraph *result;
 	int			i,
@@ -2009,17 +2007,25 @@ packGraph(TrgmNFA *trgmNFA, MemoryContext rcontext)
 	qsort(arcs, arcIndex, sizeof(TrgmPackArcInfo), packArcInfoCmp);
 
 	/* We could have duplicates because states were merged. Remove them. */
-	/* p1 is probe point, p2 is last known non-duplicate. */
-	p2 = arcs;
-	for (p1 = arcs + 1; p1 < arcs + arcIndex; p1++)
+	if (arcIndex > 1)
 	{
-		if (packArcInfoCmp(p1, p2) > 0)
+		/* p1 is probe point, p2 is last known non-duplicate. */
+		TrgmPackArcInfo *p1,
+				   *p2;
+
+		p2 = arcs;
+		for (p1 = arcs + 1; p1 < arcs + arcIndex; p1++)
 		{
-			p2++;
-			*p2 = *p1;
+			if (packArcInfoCmp(p1, p2) > 0)
+			{
+				p2++;
+				*p2 = *p1;
+			}
 		}
+		arcsCount = (p2 - arcs) + 1;
 	}
-	arcsCount = (p2 - arcs) + 1;
+	else
+		arcsCount = arcIndex;
 
 	/* Create packed representation */
 	result = (TrgmPackedGraph *)
@@ -2350,4 +2356,4 @@ printTrgmPackedGraph(TrgmPackedGraph *packedGraph, TRGM *trigrams)
 	pfree(buf.data);
 }
 
-#endif   /* TRGM_REGEXP_DEBUG */
+#endif							/* TRGM_REGEXP_DEBUG */
